# Phantom: General Trigger Attacks on Retrieval Augmented Language Generation

This repository is the official implementation of the paper: Phantom: General Trigger Attacks on Retrieval Augmented Language Generation


## Setup

Note this code will not train any models, but still requires a GPU to be run efficiently.
This setup assumes you have an NVIDIA GPU with Cuda capabilities and have installed the appropriate drivers.

Create a new virtual environment
```bash
conda create --name nlp python=3.10
conda activate nlp

pip install -r requirements.txt
```

Change the value of `base_dir` in `prj_rag/constants.py` to point to a storage location with sufficient space (at least ~200 GBs).

## Usage

**WARNING:** Running these attacks may generate text that is offensive, disrespectful, or otherwise harmful. This code should only be used in a controlled environment and with caution.

First prepare the data by encoding the dataset using the retriever model. Then run the full pipeline or individual attacks.

The main scripts are.

**Data Preparation**
```bash
python encode_dataset.py --dataset msmarco --ret_model contriever
```
Which will encode the MSMARCO dataset using the Contriever model.


**Full local attack pipeline**
```bash
python attack_local.py --config {config} --device {device} --seed {seed}
```
Which will run the full attack pipeline on the specified configuration, and save the results in `.yaml` files.

Configurations are stored in `configs/` and can be modified to run different experiments.
This folder includes the configuration files used for all the main results shown in the paper.

### Reproducing results

We provide bash scripts to reproduce our results. These scripts will run the full attack pipeline on the specified configuration and save the results in the corresponding directory in the results folder.
Below are the list of scripts for each main objective.

#### Denial of Service
```
scripts/main_table_dos_runs.sh
```

#### Biased Opinion
```
scripts/main_table_bop_runs.sh
```

#### Passage Exfiltration
```
scripts/final_privacy_gemma7b_runs.sh
scripts/final_privacy_llama3_runs.sh
scripts/final_privacy_vicuna_runs.sh
scripts/final_privacy_vicuna13b_runs.sh
```

#### Tool usage
```
scripts/email_test.sh
```

#### Harmful Behavior
```
scripts/harmful_behavior.sh
```

#### Runs without MCG
```
scripts/all_nogcg_runs.sh
```

#### Ablations
Additional scripts and config files for the ablations can be found in the respective folders.

## Acknowledgement

- Our code uses some utility functions for data loading from: https://github.com/sleeepeer/PoisonedRAG. We include that repository in this codebase.