# Orchestrates the attack on the local machine

import os
import argparse

from prj_rag import common, constants

# Attack the retriever
parser = argparse.ArgumentParser(
    description="Run the local attack on the retriever and generate prompts."
)
parser.add_argument("--config", help="Config file path", type=str, required=True)
parser.add_argument("--device", help="Device name", type=str, default="cuda:0")
parser.add_argument("--seed", help="Random seed", type=int, default=42)
parser.add_argument("--noatk", help="Skip the attack", action="store_true")
parser.add_argument(
    "--existing_ret_atk", help="Use existing retriever attack", type=str, default=""
)
args = parser.parse_args()

config = common.load_dict_from_yaml(args.config)
args = vars(args)
args.update(config)

exp_pth, _ = common.get_exp_dir(args, constants.res_dir)
print(f"\nRunning experiment: {exp_pth}\n")

# Flag to skip the attack
noatk = args.get("noatk", False)
if noatk:
    print("Skipping the attack")

# Attack the retriever using attack_retriever.py
ret_cmd = f"python attack_retriever.py --config {args['config']} --device {args['device']} --seed {args['seed']}"
existing_ret_atk = args.get("existing_ret_atk", "")
if existing_ret_atk:
    pass
else:
    if noatk:
        os.system(ret_cmd + " --noatk")
    else:
        os.system(ret_cmd)

# Attack the generator
gen_cmd = f"python attack_generator.py --config {args['config']} --device {args['device']} --seed {args['seed']}"
if existing_ret_atk:
    gen_cmd += f" --existing_ret_atk {existing_ret_atk}"
    os.system(gen_cmd)
elif noatk:
    os.system(gen_cmd + " --noatk")
else:
    os.system(gen_cmd)

# Generate Outputs
op_cmd = f"python generate_outputs.py --config {args['config']} --device {args['device']} --seed {args['seed']}"
if noatk:
    os.system(op_cmd + " --noatk")
else:
    os.system(op_cmd)

# Evaluate the outputs
os.system(f"python evaluate_outputs.py --exp_pth {exp_pth}")
